@@ -9,7 +9,10 @@ typedef enum {
     /* Unknown Perl svtype within the structure. */
     json_create_unknown_type,
     json_create_bad_char,
-    json_create_unicode_too_big,
+    /* An error from the unicode.c library. */
+    json_create_unicode_error,
+    /* A printed number turned out to be longer than MARGIN bytes. */
+    json_create_number_too_long,
 }
 json_create_status_t;
 
@@ -57,9 +60,12 @@ static int (* json_create_error_handler) (const char * file, int line_number, co
     }
 
 #define BUFSIZE 0x4000
-/* Add a margin of 16 bytes in case some stupid code reads after the
-   end of the buffer. */
-#define MARGIN 0x10
+
+/* MARGIN is the size of the "spillover" area where we can print
+   numbers or Unicode UTF-8 whole characters (runes) into the buffer
+   without having to check the printed length after each byte. */
+
+#define MARGIN 0x40
 
 typedef struct json_create {
     /* The length of the input string. */
@@ -70,6 +76,19 @@ typedef struct json_create {
 }
 json_create_t;
 
+/* Check the length of the buffer, and if we don't have more than
+   MARGIN bytes left to write into, then we put "jc->buffer" into the
+   Perl scalar "jc->output" via "json_create_buffer_fill". We always
+   want to be at least MARGIN bytes from the end of "jc->buffer" after
+   every write operation so that we always have room to put a number
+   or a UTF-8 "rune" in the buffer without checking the length
+   excessively. */
+
+#define CHECKLENGTH				\
+    if (jc->length >= BUFSIZE - MARGIN) {	\
+	CALL (json_create_buffer_fill (jc));	\
+    }
+
 /* Everything else in this file is ordered from callee at the top to
    caller at the bottom, but because of the recursion as we look at
    JSON values within arrays or hashes, we need to forward-declare
@@ -113,9 +132,7 @@ add_char (json_create_t * jc, unsigned char c)
     jc->buffer[jc->length] = c;
     jc->length++;
     /* The size we have to use before we write the buffer out. */
-    if (jc->length >= BUFSIZE - MARGIN) {
-	CALL (json_create_buffer_fill (jc));
-    }
+    CHECKLENGTH;
     return json_create_ok;
 }
 
@@ -146,10 +163,24 @@ static INLINE json_create_status_t
 add_str_len (json_create_t * jc, const char * s, unsigned int slen)
 {
     int i;
-    for (i = 0; i < slen; i++) {
-	unsigned char c;
-	c = (unsigned char) s[i];
-	CALL (add_char (jc, c));
+    /* Hopefully, the compiler optimizes the following "if" statement
+       away to a true value for almost all cases. */
+    if (slen < MARGIN) {
+	/* Gonna take you right into the DANGER ZONE. */
+	for (i = 0; i < slen; i++) {
+	    /* DANGER ZONE! */
+	    jc->buffer[jc->length + i] = s[i];
+	}
+	/* Phew. We survived the DANGER ZONE. */
+	jc->length += slen;
+	CHECKLENGTH;
+    }
+    else {
+	/* A very long string which may overflow the buffer, so use
+	   checking routines. */
+	for (i = 0; i < slen; i++) {
+	    CALL (add_char (jc, (unsigned char) s[i]));
+	}
     }
     return json_create_ok;
 }
@@ -160,19 +191,42 @@ add_str_len (json_create_t * jc, const char * s, unsigned int slen)
 
 #define ADD(x) CALL (add_str_len (jc, x, strlen (x)));
 
-/* Add a "\u3000". */
+
+static INLINE json_create_status_t
+add_one_u (json_create_t * jc, unsigned int u)
+{
+    char hex[5];
+    ADD ("\\u");
+    snprintf (hex, 4, "%04u", u);
+    CALL (add_str_len (jc, hex, 4));
+    return json_create_ok;
+}
+
+/* Add a "\u3000" or surrogate pair if necessary. */
 
 static INLINE json_create_status_t
 add_u (json_create_t * jc, unsigned int u)
 {
-    char hex[5];
-    ADD ("\\u");
     if (u > 0xffff) {
-	return json_create_unicode_too_big;
+	unsigned hi;
+	unsigned lo;
+	int status = unicode_to_surrogates (u, & hi, & lo);
+	if (status != UNICODE_OK) {
+	    if (JCEH) {
+		(*JCEH) (__FILE__, __LINE__,
+			 "Error %d making surrogate pairs from %X",
+			 status, u);
+	    }
+	    return json_create_unicode_error;
+	}
+	CALL (add_one_u (jc, hi));
+	/* Backtrace fallthrough. */
+	return add_one_u (jc, lo);
+    }
+    else {
+	/* Backtrace fallthrough. */
+	return add_one_u (jc, u);
     }
-    snprintf (hex, 4, "%04u", u);
-    CALL (add_str_len (jc, hex, 4));
-    return json_create_ok;
 }
 
 /* Add a string to the buffer with quotes around it and escapes for
@@ -183,7 +237,6 @@ static INLINE json_create_status_t
 json_create_add_key_len (json_create_t * jc, const unsigned char * key, STRLEN keylen)
 {
     int i;
-    char * istring;
     CALL (add_char (jc, '"'));
     for (i = 0; i < keylen; i++) {
 	unsigned char c;
@@ -229,7 +282,6 @@ json_create_add_key_len (json_create_t * jc, const unsigned char * key, STRLEN k
 static INLINE json_create_status_t
 json_create_add_string (json_create_t * jc, SV * input)
 {
-    int i;
     char * istring;
     STRLEN ilength;
     istring = SvPV (input, ilength);
@@ -241,12 +293,60 @@ static INLINE json_create_status_t
 json_create_add_integer (json_create_t * jc, SV * sv)
 {
     long int iv;
-    STRLEN ivlen;
-    char ivbuf[0x40];
+    int ivlen;
     iv = SvIV (sv);
-    ivlen = snprintf (ivbuf, 0x40, "%ld", iv);
-    /* Backtrace fall through, remember to check the caller's line. */
-    return add_str_len (jc, ivbuf, ivlen);
+
+    /* Souped-up integer printing for small integers. The following is
+       all just souped up versions of snprintf ("%d", iv);. */
+
+    if (iv < 0) {
+	CALL (add_char (jc, '-'));
+	iv = -iv;
+    }
+    if (iv < 10) {
+	CALL (add_char (jc, iv + '0'));
+    }
+    else if (iv < 100) {
+	char ivc[2];
+	ivc[0] = iv / 10 + '0';
+	ivc[1] = iv % 10 + '0';
+	CALL (add_str_len (jc, ivc, 2));
+    }
+    else if (iv < 1000) {
+	char ivc[3];
+	ivc[0] = iv / 100 + '0';
+	ivc[1] = (iv / 10) % 10 + '0';
+	ivc[2] = iv % 10 + '0';
+	CALL (add_str_len (jc, ivc, 3));
+    }
+    else if (iv < 10000) {
+	char ivc[4];
+	ivc[0] = iv / 1000 + '0';
+	ivc[1] = (iv / 100) % 10 + '0';
+	ivc[2] = (iv / 10) % 10 + '0';
+	ivc[3] = iv % 10 + '0';
+	CALL (add_str_len (jc, ivc, 4));
+    }
+    else if (iv < 100000) {
+	char ivc[5];
+	ivc[0] = iv / 10000 + '0';
+	ivc[1] = (iv / 1000) % 10 + '0';
+	ivc[2] = (iv / 100) % 10 + '0';
+	ivc[3] = (iv / 10) % 10 + '0';
+	ivc[4] = iv % 10 + '0';
+	CALL (add_str_len (jc, ivc, 5));
+    }
+    else {
+	/* The number is 100,000 or more, so we're just going to print
+	   it into "jc->buffer" with snprintf. */
+	ivlen = snprintf ((char *) jc->buffer + jc->length, MARGIN, "%ld", iv);
+	if (ivlen >= MARGIN) {
+	    return json_create_number_too_long;
+	}
+	jc->length += ivlen;
+	CHECKLENGTH;
+    }
+    return json_create_ok;
 }
 
 static INLINE json_create_status_t
@@ -254,11 +354,15 @@ json_create_add_float (json_create_t * jc, SV * sv)
 {
     double fv;
     STRLEN fvlen;
-    char fvbuf[0x40];
+    char fvbuf[MARGIN];
     fv = SvNV (sv);
-    fvlen = snprintf (fvbuf, 0x40, "%g", fv);
-    /* Backtrace fall through, remember to check the caller's line. */
-    return add_str_len (jc, fvbuf, fvlen);
+    fvlen = snprintf ((char *) jc->buffer + jc->length, MARGIN, "%g", fv);
+    if (fvlen >= MARGIN) {
+	return json_create_number_too_long;
+    }
+    jc->length += fvlen;
+    CHECKLENGTH;
+    return json_create_ok;
 }
 
 /* Add a number which is already stringified. This bypasses snprintf
